/*! \file The PHOTOSENSOR module is responsible for the management 
 *        and reading of the photosensors
 */

#ifndef PHOTOSENSOR_H_
#define PHOTOSENSOR_H_

/*! This function initializes the photosensor module
 *  and configures interrupt behaviour
 */
void PHOTOSENSOR_Init();

/*! This function enables the detection of the projectile that
 *  enters a photosensor
 */
void PHOTOSENSOR_EnableRisingEdge();

/*! This function disables the detection of the projectile that
 *  enters a photosensor
 */
void PHOTOSENSOR_DisableRisingEdge();

/*! This function enables the detection of the projectile that
 * leaves a photosensor
 */
void PHOTOSENSOR_EnableFallingEdge();

/*! This function disables the detection of the projectile that
 * leaves a photosensor
 */
void PHOTOSENSOR_DisableFallingEdge();

/*! This function disables the detection of the projectile that
 * enters or leaves a photosensor
 */
void PHOTOSENSOR_DisableBothEdges();

/*! This function gets the current status of a sensor
 *  
 *  \param sensorno The number of the sensor that has to be checked
 *  \return True if the sensorno sensor is interrupted
 */
bool PHOTOSENSOR_Status(char sensorno);

/*! This function get the number of the actual interrupted sensor
 *  
 *  \return The number of the sensor that is interrupted or -1 if no sensor
 *          is interrupted
 */
char PHOTOSENSOR_NoInterrupted();

/*! This function masks all interrupts internally.
 *  The callback function won't be called then
 */
void PHOTOSENSOR_MaskAll();

/*! This function unmasks all interrupts internally.
 *  The callback function will be called on every interrupt that is not disabled
 *  by having disabled specific edges
 */
void PHOTOSENSOR_UnmaskAll();

/*! This function masks a specific sensor
 *  
 *  \param sensorno The sensorno that has to be masked
 */
void PHOTOSENSOR_Mask(char sensorno);

/*! This function masks all interrupts internally, with the exeception of 
 *  on sensor, which will be unmasked
 *  
 *  \param sensorno The only sensorno of which we want to reach callbacks
 */
void PHOTOSENSOR_MaskAllButOne(char sensorno);

/*! This function unmasks a specific sensor
 *  
 *  \param sensorno The sensorno that has to be unmaksed
 */
void PHOTOSENSOR_Unmask(char sensorno);

/*! This function checks whether a given sensor is masked or not
 *  
 *  \param sensorno The sensorno that has to be checked
 */
bool PHOTOSENSOR_IsMasked(char sensorno);

// callback function type
// bool: true if rising edge
// char: sensorno
typedef void (*PHOTOSENSOR_callback_t)(bool, char);

/*! This function registers an error callback in the PHOTOSENSOR module
 *  that will be called when an error is caught.
 *  
 *  \param func The error callback function pointer 
 */
void PHOTOSENSOR_SetErrorCallback(PHOTOSENSOR_callback_t func);

/*! This function configures below which sensorno the errorCallback will be called
 *  
 *  \param sensorno The number of sensor below which it is guaranteed that an error occured
 *         if an interrupt has been caught
 */
void PHOTOSENSOR_SetErrorOnBelowlback(char sensorno);

/*! This handler is the interrupt handler for the Photosensor module.
 *  Every interrupt on the photosensors must be forwarded to this handler.
 */
void PHOTOSENSOR_Interrupt();

void PHOTOSENSOR_ErrorOnBelow(char sensorno);
#endif /*PHOTOSENSOR_H_*/
