// $Id$
// Created:   Apr 11, 2006 lohmann

/**
 * \file CGU.h
 * \brief Access classes for the TC1796 Clock Generation Unit (CGU)
 */

#ifndef __HW_DEV_CGU_H__
#define __HW_DEV_CGU_H__

#include "stdinc.h"

// Include trigcc hw-support headers
#include <tc1796b/scu.h>

namespace hw { namespace dev {
/**
* @class CGU
* @brief TC1796 Clock Generation Unit
* 
* @note While the TC1796 CGU can be reprogrammed at runtime, this is currently not supported.
* All clock settings are configured with pure::variants and provided as enum Consts (see below). 
* The accessor functions getCPUClockFreq() and getSysClockFreq() just return this constant values.
* You should, however, prefer using these functions over the constants, as the functions may one
* day be adviced to provide runtime adaptability of the clock rates.
*/
class CGU {
	static CGU theCGU_;
	//! Performs start-time initializations. Only called by advice.
	
	// TODO: move to advice, make private
	// currently init call is hard-coded in __main() (file triboard1796.cpp)
  public:
	void init();
	
  protected:
    CGU(){}
  public:	
	//! Returns the one-and-only CGU instance
	static CGU& Inst() {
		return theCGU_;
	}

	//! Integer type for frequency values
	typedef UInt32 Frequency;
	
	//! Returns the frequency of the CPU clock signal
	static Frequency getCPUClockFreq() {
		return CPUFREQ;
	}
	
	//! Returns the frequency of the system clock signal (SYS)
	static Frequency getSysClockFreq() {
		return SYSFREQ;	
	}
		
	//! returns the actual (runtime examined) frequency of the CPU clock signal
	static Frequency getRealCPUClockFreq();
	
	//! returns the actual (runtime examined) frequency  of the system clock signal
	static Frequency getRealSysClockFreq() {
		if( PLL_CLC.bits.SYSFS == 0 )
			return getRealCPUClockFreq() / 2;
		return getRealCPUClockFreq();
	}

	/**
	 * @brief Default CPU and system clock (SYS) settings.
	 * 
	 * @note Clock settings are configured with pure::variants. 
	 */
	enum DOXY_SYMBOL(Consts) {
		OSCFREQ			= cfHW_OSCFREQ,				//!< Oscillator frequency (20 MHz on TriBoard1796)
		PLL_NDIV		= cfHW_PLL_NDIV,			//!< PLL N-divider: 20-100			
		PLL_KDIV		= cfHW_PLL_KDIV,			//!< PLL K-divider: 1-16
		PLL_PDIV		= cfHW_PLL_PDIV,			//!< PLL P-divider: 1-8
		PLL_SYSDIV		= cfHW_PLL_SYSDIV,			//!< PLL system clock rate divider: 1-2
		CPUFREQ 		= (PLL_NDIV * OSCFREQ) / (PLL_PDIV * PLL_KDIV),	//!< resulting CPU clock rate: (OSC * N) / (P * K)
		SYSFREQ			= CPUFREQ / PLL_SYSDIV		//!< resulting SYS clock rate 
	};
};


}} // namespace hw::dev


#endif // __HW_DEV_CGU_H__
