// $Id: GPIO.h 1323 2010-12-21 15:17:02Z mike $
// Created:   05.06.2006 lohmann

/**
 * @file GPIO.h
 * @brief Contains abstractions for the TC1796 General Purpose IO (GPIO) lines
 */
 
#ifndef HW_DEV_GPIO_H_
#define HW_DEV_GPIO_H_

#include "stdinc.h"
#include "hw/triboard1796.h"

	#include "hw/tricore-abs-noabs-FIX.h"
#if __GNUC__ >=3 || __TOOL_VERSION__ >= 210
	#include <tc1796b/p0_p1_p2_p3_p4_p5_p6_p7_p8_p9/sharetypes.h>
#else
	#include <tc1796b/p0_p1_p10_p2_p3_p4_p5_p6_p7_p8_p9/sharetypes.h>
#endif
	#include <tc1796b/p2/types.h>
	#include <tc1796b/p5/types.h>
	#include <tc1796b/p6/types.h>
	#include <tc1796b/p8/types.h>
	#include <tc1796b/p9/types.h>


namespace hw { namespace _tc {
	//! Internal memory layout of a TC1796 GPIO port
	struct PORT_t {
		Pn_OUT_t_nonv		OUT;		//!< Port 0-9 output register. 					
		Pn_OMR_t_nonv		OMR;		//!< Port 0-9 output modification register. 			
		unsigned int		res1[2];	
		Pn_IOCRm_t_nonv		IOCR[4];	//!< Port 0-9 IO control registers (for bits 0-15).
		unsigned int		res2;	
		Pn_IN_t_nonv		IN;			//!< Port 0-9 input register. 	
		unsigned int		res3[6];	
		union {
			Pn_PDR_t_nonv	PDR;		//!< Port 0,1,3,4 pad driver mode register.		
			P2_PDR_t_nonv	P2_PDR;		//!< Port 2 pad driver mode register.	
			P5_PDR_t_nonv	P5_PDR;		//!< Port 5 pad driver mode register.	
			P6_PDR_t_nonv	P6_PDR;		//!< Port 6 pad driver mode register.	
			P8_PDR_t_nonv	P8_PDR;		//!< Port 8 pad driver mode register.	
			P9_PDR_t_nonv	P9_PDR;		//!< Port 9 pad driver mode register.	
	   };
		unsigned int		res4[3];
		union {
			Pn_ESR_t_nonv	ESR;		//!< Port 3,4,8,9 GPTA emergency stop register.	
			P2_ESR_t_nonv	P2_ESR;		//!< Port 2 GPTA emergency stop register.	
		};	  
	};

}}	// namespace hw::_tc


namespace hw { namespace dev {

/**
 * @class PortBase
 * @brief TC1796 common base for the GPIO classes
 * 
 */
template< int ID_, int BITS_, unsigned MASK_>
class PortBase {
  protected:
	//! We only serve as a base class
	PortBase() {};
  
  public:

 	enum DOXY_SYMBOL(Consts) {
  		ID 		= ID_,			//!< Identifier of this port (hw port number)
  		BITS	= BITS_,		//!< Number of available bits
  		MASK	= MASK_			//!< Mask of available bits
  	};
 
	//! Provides access to the actual hardware register set
	static hw::_tc::PORT_t& port() {
		const UInt32 BASE = 0xf0000c00;	// Port 0 base address
		const UInt32 SIZE = 0x100;		// Register window size (=offset to next port)
		
		return *reinterpret_cast< hw::_tc::PORT_t* > ( BASE + ID * SIZE );
	}
	
	/**
	 * @brief Mode settings for a GPIO pin
	 */ 
	enum DOXY_SYMBOL(Mode) {
		INPUT				=	0x0,	//!< Use as general input (default after reset)
		OUTPUT				=	0x8,	//!< Use as general output
	
		OUTPUT_PUSHPULL 	=	0x8,	//!< Use as general output with push-pull driver 
		OUTPUT_OPENDRAIN	=	0xC,	//!< Use as general output with open drain

		OUTPUT_ALT1			= 	0x1,	//!< Activate alternative function 1
		OUTPUT_ALT2			= 	0x2,	//!< Activate alternative function 2
		OUTPUT_ALT3			=	0x3,	//!< Activate alternative function 3

		INPUT_PULLDOWN		=	0x1,	//!< Use as input, internal pull down driver connected
		INPUT_PULLUP		=   0x2,	//!< Use as input, internal pull up driver connected
		INPUT_NOPULL		= 	0x3		//!< Use as input, no internal pull driver connected
	};	

	//!@name Low-Level Register Access
	//!@{
	
	static inline Pn_OUT_t_nonv& OUT() {
		return port().OUT;
	};

	static inline Pn_OMR_t_nonv& OMR() {
		return port().OMR;
	};

	static inline Pn_IOCRm_t_nonv& IOCR(int m) {
		return port().IOCR[ m ];
	};

	static inline Pn_IN_t_nonv IN() {
		return port().IN;
	};

	static inline Pn_PDR_t_nonv PDR() {
		return port().PDR;
	};

	static inline Pn_ESR_t_nonv ESR() {
		return port().ESR;
	};
	
	//!@}
	
	//!@name Bit-wise Access
	//!@{

	static void set( unsigned bit ) {
		OMR().reg = ( 1 << bit );
	};
	static void clear( unsigned bit ) {
		OMR().reg = ( 1 << ( bit + 16 ) );
	};
	static void toggle( unsigned bit ) {
		bit = 1 << bit;
		OMR().reg = bit | bit << 16;
	};
	static volatile unsigned get( unsigned bit ) {
		return IN().reg & (1 << bit );
	};
	static void setMode( unsigned bit, unsigned mode ) {
		// ensure byte-wise access
		unsigned char* addr = reinterpret_cast< unsigned char* >( &port().IOCR[0].reg ) + bit;
		*addr =  mode << 4;  
	}
	//!@}
};

}} // namespace hw::dev
#endif /*HW_DEV_GPIO_H_*/
