// $Id: IRQCtrl.h 1211 2010-07-16 13:05:25Z sijesche $
// Created:   25.03.2006 lohmann

/**
 * \file IRQCtrl.h
 * \brief Access classes for TriCore service request nodes (SRNs).
 *
 * On TriCore, all devices that may trigger IRQs provide a SRN-register for each IRQ 
 * they may trigger, which is used to configure the IRQL, trigger an IRQ or reset
 * a service request. As all SRN-registers use the same standard layout (with some
 * optional device-dependend bits), they can be handled by a single template class IRQCtrl. 
 */

#ifndef __HW_DEV_IRQCTRL_H__
#define __HW_DEV_IRQCTRL_H__

#include "stdinc.h"
#include "util/types.h"

/**
 * @class SRC_t_nonv
 * @brief Common format of a TriCore service request node register (SRN)
 *
 */

typedef union { 
   struct {
      unsigned int SRPN:8;        	//!< service request priority number
      unsigned int _bit8:1;              
      unsigned int _bit9:1;             
      unsigned int TOS:1;           //!< type of service (0 = CPU, 1 = PCP)    
      unsigned int _bit11:1;            
      unsigned int SRE:1;           //!< interrupt enabled    
      unsigned int SRR:1;           //!< service request pending    
      unsigned int CLRR:1;          //!< set SRR ( =1 sets SRR bit)    
      unsigned int SETR:1;          //!< clear SRR ( =1 clears SRR bit)    
	  unsigned int _bit16_31:16;
   } bits;
   unsigned long reg; 
} CIAO_ALIGNED(4) SRC_t_nonv;


namespace hw { namespace dev {

/**
 * \class IRQCtrl
 * \brief An IRQCtrl encapsulates the control of a single IRQ line. It
 * \brief provides means to enable/disable or test the IRQ line.
 * 
 * On TriCore, almost all interrupt service control registers use the same
 * layout. Therefore the default type for REG_T is suitable in most cases.
 */
template< UIntPtr REG_ADDR, typename REG_T = SRC_t_nonv > 
struct IRQCtrl {
	typedef REG_T Register;
	enum DOXY_SYMBOL(Consts){ REGISTER = REG_ADDR }; 
	//! Return register value 
	static Register& reg() {
		return *( (Register*) (REGISTER) );
	}
	
	//! Unmasks IRQ, so that it takes part in the IRQ arbitration process. 
	static void enable() {
		reg().bits.SRE = 1;
	}
	//! Masks IRQ, so that it does not take part in the IRQ arbitration process. 
	//
	//! @note Disableing thie IRQ usually does not disable the device itself.
	static void disable() {
		reg().bits.SRE = 0;
	}
	//! Returns true, if IRQ is enabled
	static Bool IE() {
		return reg().bits.SRE;
	}
	
	//! Returns true, if service request is pending. 
	static Bool pending() {
		return reg().bits.SRR;
	}
	//! Explicitly set service request pending flag.
	static void set() {
		reg().bits.SETR = 1;
	}
	//! Explicitly clear service request pending flag.
	static void clear() {
		reg().bits.CLRR = 1;
	}
	
	//! Returns the IRQ priority number (=vector number) requested by this IRQ
	static UInt8 IRQL() {
		return reg().bits.SRPN;
	}
	//! Sets the IRQ priority number (=vector number) of this IRQ
	static void IRQL( UInt8 level ) {
		reg().bits.SRPN = level;
	}
};

}} // namespace hw::dev

#endif // __HW_DEV_IRQCTRL_H__
