/*
 * VirtioNet.h
 *
 *  Created on: Mar 8, 2010
 *      Author: bossel00
 */

#ifndef VIRTIONET_H_
#define VIRTIONET_H_

#include "stdinc.h"
#include "VirtioPCI.h"

#include <stdio.h>

namespace hw {
namespace dev {

class VirtioNet : public VirtioPCI {

public:
	struct ethernet_packet {
		UInt8  dst[6];
		UInt8  src[6];
		UInt16 type;
		char   data[1500];
		//UInt32 crc;
	};

private:

	// structs as seen in paper "virtio: towards a de-facto standard for virtual I/O devices",
	//                          ACM SIGOPS Operating Systems Review, Volume 42, Issue 5
	// http://portal.acm.org/citation.cfm?id=1400097.1400108

	struct vring_desc {
		enum {
			VRING_DESC_F_NEXT = 1,
			VRING_DESC_F_WRITE = 2
		};

		UInt64 addr;
		UInt32 len;
		UInt16 flags;
		UInt16 next;
	};

	struct vring_avail {
		UInt16 flags;
		UInt16 idx;
		UInt16 ring[];
	};

	struct vring_used_elem {
		UInt32 id;
		UInt32 len;
	};

	struct vring_used {
		UInt16 flags;
		UInt16 idx;
		vring_used_elem ring[];
	};

	struct virtio_net_hdr {

		enum {
			VIRTIO_NET_HDR_F_NEEDS_CSUM = 1
		};

		enum {
			VIRTIO_NET_HDR_GSO_NONE  = 0,
			VIRTIO_NET_HDR_GSO_TCPV4 = 1,
			VIRTIO_NET_HDR_GSO_UDP   = 3,
			VIRTIO_NET_HDR_GSO_TCPV6 = 4,
			VIRTIO_NET_HDR_GSO_ECN   = 0x80
		};

		UInt8 flags;
		UInt8 gso_type;
		UInt16 hdr_len;
		UInt16 gso_size;
		UInt16 csum_start;
		UInt16 csum_offset;
	};

	// feature bits
	enum {
		VIRTIO_NET_F_CSUM = 0,
		VIRTIO_NET_F_GUEST_CSUM = 1,
		VIRTIO_NET_F_MAC = 5,
		VIRTIO_NET_F_GSO = 6,
		VIRTIO_NET_F_GUEST_TSO4 = 7,
		VIRTIO_NET_F_GUEST_TSO6 = 8,
		VIRTIO_NET_F_GUEST_ECN = 9,
		VIRTIO_NET_F_GUEST_UFO = 10,
		VIRTIO_NET_F_HOST_TSO4 = 11,
		VIRTIO_NET_F_HOST_TSO6 = 12,
		VIRTIO_NET_F_HOST_ECN = 13,
		VIRTIO_NET_F_HOST_UFO = 14
	};

	enum QueueSelector {
		VIRTIO_RECEIVE_QUEUE_SELECTOR = 0,
		VIRTIO_TRANSMIT_QUEUE_SELECTOR = 1
	};

public:

private:

	static VirtioNet inst_;

	UInt8 vector;

	static const int RING_SIZE = 0x100;

	// vring descriptors (aligned on page boundary):
	//    0x0 - 0x0FFF: vring_desc
	// 0x1000 - 0x1FFF: vring_avail
	// 0x2000 - 0x2FFF: vring_used
	//          0x3000: vring_end
	struct vring {
		volatile vring_desc    desc[RING_SIZE] __attribute__((aligned(4096)));  // 4096 Bytes
		volatile vring_avail   avail           __attribute__((aligned(4096)));  // 4096 Bytes
		volatile vring_used    used            __attribute__((aligned(4096)));  // 4096 Bytes
		unsigned long          end[0]          __attribute__((aligned(4096)));  //    0 Bytes

		UInt16        last_used;

		UInt16        next_free;
		UInt16        next_avail;

		// common helper functions
		int  get_next_desc();
		void add_desc(int desc);

		// returns true iff no unused descriptors available
		bool empty() {
			return (UInt16(last_used - used.idx) % RING_SIZE > 126U);
		}

	} transmit_ring_, receive_ring_;


	// trigger host
	void notify(QueueSelector queue) {
		io_.vring_queue_notifier.outw(queue);
	}

	virtio_net_hdr net_hdr_none;

	// TODO: Puffer in Empfangs-Slice auslagern
	volatile struct {
		virtio_net_hdr  net_hdr;
		ethernet_packet packet;
	} receive_buffer_[RING_SIZE/2];

public:
	static VirtioNet& Inst() { return inst_; }

	void init();

	const char* getName() { return "virtio paravirtualized network adapter"; }

	UInt64 getDefaultAddress();

	UInt8 getVector() { return vector; }

	// gets and clears interrupt status register
	UInt8 getISR() { return io_.pci_isr.inb(); }

	// wait until all unprocessed packets are sent
	void flush() {
		while (transmit_ring_.used.idx != transmit_ring_.last_used);
	}

	// TODO: send() und receive() in slices auslagern
	// flush() must be called before freeing data
	void send(const void* data, unsigned datasize);

	bool hasBeenSent(const void* data);

	void send(UInt64 dst, UInt16 type, const void* payload, unsigned payloadsize) {
		// TODO: Daten kopieren und senden
	}

	unsigned receive(void* data, unsigned datasize);

	bool packet_received() {
		//return receive_ring_.used.idx != receive_ring_.last_used;
		return static_cast<UInt16>(receive_ring_.last_used - receive_ring_.used.idx) % RING_SIZE != RING_SIZE/2-1;
	}

	const void* nextData();
	unsigned    nextDataLength();
	void        nextFree();

	void debug() {
		printf("initial state:\n");
		printf(" Host-Features: %x\n", io_.host_features.inl());
		printf(" Guest-Features: %x\n", io_.guest_features.inl());
		printf(" VRing-PageNum: %x\n", io_.vring_page_num.inl());
		printf(" VRing-RingSize: %x\n", io_.vring_ring_size.inw());
		printf(" VRing-Queue-Selector: %x\n", io_.vring_queue_selector.inw());
		printf(" VRing-Queue-Notifier: %x\n", io_.vring_queue_notifier.inw());
		printf(" Status: %x\n", io_.status.inb());
		printf(" PCI-ISR: %x\n", io_.pci_isr.inb());

		printf(" Config: %x %x %x %x %x %x\n", io_.config.offset(0).inb(), io_.config.offset(1).inb(), io_.config.offset(2).inb(),
				io_.config.offset(3).inb(), io_.config.offset(4).inb(), io_.config.offset(5).inb());

		printf("-----------------------------------------\n");



		const int NUM_PACKETS = 3;
	     static ethernet_packet packet[256];

		for (int i = 0; i < 256; ++i) {
			packet[i].dst[0] = 0x12;
			packet[i].dst[1] = 0x34;
			packet[i].dst[2] = 0x56;
			packet[i].dst[3] = 0x78;
			packet[i].dst[4] = 0x9A;
			packet[i].dst[5] = 0xBC;

			packet[i].src[0] = 0x1A;
			packet[i].src[1] = 0xB2;
			packet[i].src[2] = 0x3C;
			packet[i].src[3] = 0xD4;
			packet[i].src[4] = 0x5E;
			packet[i].src[5] = 0xF6;

			packet[i].type = 0x6308;
		}


		unsigned counter = 0;

		//for (;;)
		for (int i = 0; i < NUM_PACKETS; ++i) {
			//for (volatile int j = 0; j < 1000; ++j);  // ein bisschen warten
			int id = transmit_ring_.get_next_desc();

			transmit_ring_.desc[id].addr = (unsigned long) &net_hdr_none;
			transmit_ring_.desc[id].len = sizeof(net_hdr_none);
			transmit_ring_.desc[id].flags = vring_desc::VRING_DESC_F_NEXT;

			transmit_ring_.desc[id + 1].addr = (unsigned long) &packet[i % 256];
			transmit_ring_.desc[id + 1].len = sizeof(ethernet_packet);
			transmit_ring_.desc[id + 1].flags = 0;

			packet[i % 256].data[0] = counter >> 24;
			packet[i % 256].data[1] = (counter >> 16) & 0xFF;
			packet[i % 256].data[2] = (counter >> 8) & 0xFF;
			packet[i % 256].data[3] = counter++ & 0xFF;

			transmit_ring_.add_desc(id); // senden
			notify(VIRTIO_TRANSMIT_QUEUE_SELECTOR);
		}

		printf("Senden fertig...\n");

		for (;;) {
			ethernet_packet packet;
			receive(&packet, sizeof(packet));
			printf("Paket empfangen: %x %x\n", (unsigned int) packet.data[10], (unsigned int) packet.data[11]);
		}

		printf("-----------------------------------------\n");
		printf("VRing-PageNum: %x\n", io_.vring_page_num.inl());
		printf(" Host-Features: %x\n", io_.host_features.inl());
		printf(" Guest-Features: %x\n", io_.guest_features.inl());
		printf(" VRing-PageNum: %x\n", io_.vring_page_num.inl());
		printf(" VRing-RingSize: %x\n", io_.vring_ring_size.inw());
		printf(" VRing-Queue-Selector: %x\n", io_.vring_queue_selector.inw());
		printf(" VRing-Queue-Notifier: %x\n", io_.vring_queue_notifier.inw());
		printf(" Status: %x\n", io_.status.inb());
		printf(" PCI-ISR: %x\n", io_.pci_isr.inb());
		//printf("-----------------------------------------\n");

		printf("transmit_ring_.used.idx = %d, transmit_ring_.used.flags = %x\n", transmit_ring_.used.idx, transmit_ring_.used.flags);
		printf("transmit_ring_.used.ring[0].id = %d, transmit_ring_.used.ring[0].len = %d\n", transmit_ring_.used.ring[0].id, transmit_ring_.used.ring[0].len);

		flush();

		printf("Debug Ende...\n");
		for (;;);
	}
};

}
}

#endif /* VIRTIONET_H_ */
