/** \author Christian Meier
 *  \version $Id: Controller.cpp 995 2009-08-11 14:26:31Z mike $
 */

#include "InterruptScheme.h"
#include "Controller.h"

namespace hw { namespace dev { namespace can { namespace multican {

Controller &
Controller::Inst(ControllerIdType ControllerId) {
	return InstByUnit(CtrlIdToCtrlUnit(ControllerId));
}

ReturnType
Controller::InitController(ControllerConfigType const* p_Config) {
	{ // NCR
		CAN_NCRm_t_nonv ncr;
		ncr.reg = 0;

		ncr.bits.INIT	= 1;			// CANIF_CS_STOPPED implicit req: CAN259
		ncr.bits.TRIE	= 0;			// Transfer interrupt enable; we don't use this interrupt
		ncr.bits.LECIE  = 0;			// Last Error Code interrupt enable; we don't use this interrupt
		ncr.bits.ALIE	= 1;			// Alert interrupt, needed for Busoff interrupt (in polling mode this is don't care
										// As we enable interrupts through the InterruptManager separatly
		ncr.bits.CCE	= 1;			// Configuration Change Enable: For now enable it until configuration is made
		ncr.bits.CANDIS = 0;			// We don't want to disable it as we can asume that it is already disabled (req: CAN256)
		ncr.bits.CALM	= 0;			// We don't use Analyze mode
		ncr.bits.SUSEN  = 0;			// Susend enable Currently not used

		// Commit to hardware
		_NCR.reg = ncr.reg;
	}

	{ // NSR
		// It is a good idea to "acknowledge" all old Interrupt request flags here
		_NSR.reg = 0;
	}

	{ // NIPR
		CAN_NIPRm_t_nonv nipr;

		nipr.bits.ALINP		= HW_DEV_CAN_MULTICAN_ALERT_INT(p_Config->m_ControllerId);			// Set according to controller number
		nipr.bits.CFCINP	= 0xf;		// We don't use frame counter interrupt
		nipr.bits.TRINP		= 0xf;		// We don't use transfer interrupt
		nipr.bits.LECINP	= 0xf;		// We don't use Last Error code Interrupt

		_NIPR.reg = nipr.reg;
	}

	{ // NPCR
		CAN_NPCRm_t_nonv npcr;
		npcr.reg	= 0;

		npcr.bits.RXSEL	= 0;		// We currently do not allow any configuration of the output pins so
									// a specific node id represents a specific pin pair of the tricore, BUT:
		npcr.bits.LBM = InternalConnectConfig(p_Config->m_ControllerId);

		// Commit to hardware
		_NPCR.reg = npcr.reg;
	}

	{ // NBTR
		CAN_NBTRm_t_nonv nbtr;
		nbtr.reg	= 0;

		nbtr.bits.BRP	= BRPConfig(p_Config->m_ControllerId);
		nbtr.bits.DIV8	= DIV8Config(p_Config->m_ControllerId);

		nbtr.bits.SJW	= 0;		// For now we just use 1 time quantum as jump width --> SJW = 0 (CHECK)

		// The multican module does not differentiate between Propagation Segment and Phase Segment 1.
		// Instead it is nesseccary to configure the sum of Propagation Segment and Phase Segment 1 in TSEG1.
		nbtr.bits.TSEG1	= p_Config->m_PropSeg + p_Config->m_PhaseSeg1 + 1;
		nbtr.bits.TSEG2	= p_Config->m_PhaseSeg2 + 1;

		nbtr.bits.FTX	= 0;		// Fast transmit; We do not use Time Triggered Can

		// Commit to hardware
		_NBTR.reg = nbtr.reg;
	}

	{ // NECNT
		CAN_NECNTm_t_nonv necnt;
		necnt.reg = 0;

		necnt.bits.REC		= 0;		// We reset the receive error counter
		necnt.bits.TEC		= 0;		// We reset the transmit error counter
		necnt.bits.EWRNLVL	= 0xFF;		// Our design doesn't need a specific Warning Interrupt but the line
										// Is shared with Busoff interrupt, so we set EWRNLVL to the highest possible value

		// Commit to hardware
		_NECNT.reg = necnt.reg;
	}

	_NCR.bits.CCE = 0;			// Now we can disable Configuration change enable

	{ // NFCR
		// We just write reset value here
		_NFCR.reg = 0;
	}

	return CAN_OK;
}

ReturnType
Controller::SetControllerMode (StateTransitionType Transition) {
	switch(Transition) {
	case CAN_T_SLEEP:
		// There is no sleep mechanism in hardware so we simply use CAN_T_STOP	req: 	CAN258
	case CAN_T_STOP:
		// TODO: Cancel pending messages: req: CAN282
		_NCR.bits.INIT	= 1;
		break;
	case CAN_T_WAKEUP:
	case CAN_T_START:
		_NCR.bits.INIT	= 0;
	}

	return CAN_OK;
}

ControllerIdType Controller::ControllerMultiplicity_ = 0;

}}}}
