#ifndef STMTIMER1_H_
#define STMTIMER1_H_


#include "STMTimerBase.h"
#include "CPU.h"

namespace hw { namespace hal {

using namespace hw::dev;

/**
 * @brief typically used to implement hal::SystemTimer
 *
 * @tc STM timers are build on top of the system timer module (STM). STMTimer1 is hard-wired
 * to use the CMP1 compare-and-match register and to use hw::irq::STM_SRC1 as CPU interrupt SRN.
 *
 */
class STMTimer1 : public STMTimerBase< hw::dev::STM::IRQ_SRC1 > {
	static STMTimer1 theSTMTimer1_;

	//! Performs start-time initializations. Only called by advice.
	void init();

  protected:

    STMTimer1(){}

  public:

  	typedef STMTimerBase< hw::dev::STM::IRQ_SRC0 > Super;

	//! Returns the one-and-only Timer0 instance
	static STMTimer1& Inst() {
		return theSTMTimer1_;
	}


	/**
	 * @brief  Windup timer as single shot
	 *
	 * A single shot timer triggers the timing protection handling
	 */
	CIAO_STD void windupAndStart( Unit time );

	CIAO_STD void stop() {
		hw::dev::STM::ICR().bits.CMP1EN = 0;
	}

	Unit getTimer1(){
		return hw::dev::STM::TIM<0>(); //there is only one hardware clock, hence TIM<0>
	}

	CIAO_STD static Bool running() {
		return hw::dev::STM::ICR().reg & STM_ICR_CMP1EN_MASK;
	}

	CIAO_STD static void clear() {
		hw::dev::STM::ISRR().bits.CMP1IRR = 1;
		_dsync();			// ensure bit has been cleared (or do we need _isync() here?)
		IRQ::clear();		// also clear (possibly already pending) SRN flag
	}

	//! Returns true, iff service request is pending.
	static Bool pending() {
		// test for pending internal IRQ should be enough,
		// as SRN flag should always be in line with it.
		return hw::dev::STM::ICR().reg & STM_ICR_CMP1IR_MASK;
	}

	//! Explicitly set service request pending flag.
	static void set() {
		// set internal IRQ flag, which automatically sets the SRN flag, iff enabled
		hw::dev::STM::ISRR().bits.CMP1IRS = 1;
	}

	/**
	 * @brief Acknowledge timer event. To be called by the IRQ handler (prologue!).
	 *
	 * @todo Maybe we should provide an aspect for this.
	 */
	CIAO_STD void CIAO_INLINE irqAck() {
		stop();

		// internal pending bit has to be cleared explicitly,
		// SRN pending bit has already been cleared by hardware
		hw::dev::STM::ISRR().bits.CMP1IRR = 1;
	}

};

}} //namespaces
#endif /*STMTIMER1_H_*/
