#ifndef _OS_KRN_MESSAGEPORT_H_
#define _OS_KRN_MESSAGEPORT_H_

/*****************************************************************************
 * Message Pools, Sender and Receiver Ports all need be global symbols to
 * allow us moving them to specific memory regions at link time.
 * ***************************************************************************/

/* Example Usage:
 * Note: the ports are now generated from config.xml

int mpool[20];

extern os::krn::ReceiverPort<int, 20, mpool> receiver;

os::krn::SenderPort<int, 20, mpool> sender = {
	&receiver, 0,
};

os::krn::ReceiverPort<int, 20, mpool> receiver = {
	&sender, 0,
};

*/ 


namespace os { namespace krn {
	template <typename MT, int NM> class ReceiverPort;

	template <typename MSGTYPE, int msgCount>
		class SenderPort {
	
		public:
		ReceiverPort<MSGTYPE, msgCount> *receiver;
		
		// index of the next message to allocate in the pool
		int writerIndex;

		// the Message Pool
		MSGTYPE *pool;

		// basic primitives
		MSGTYPE *allocateMessage();
		void sendMessage(MSGTYPE *message);
		};

		template <typename MSGTYPE, int msgCount>
		class ReceiverPort {

		public:
		// the connected sender port
		SenderPort<MSGTYPE, msgCount> *sender;

		// index of the next message to release
		int readerIndex;

		// basic primitives
		MSGTYPE *receiveMessage();
		void releaseMessage(MSGTYPE *message);
		inline bool peek() { 
			return (readerIndex != sender->writerIndex);
		}
	};

	template <typename MSGTYPE, int msgCount>
		class ReceiverPortWriteAccess : public ReceiverPort<MSGTYPE, msgCount> {

		};

	/**************************************************************************/
	/**********************        Implementation        **********************/
	/**************************************************************************/
	template <typename MSGTYPE, int msgCount>
		MSGTYPE *SenderPort<MSGTYPE, msgCount>::allocateMessage() {
			int pos;

			// buffer full?
			// reader and writer point to the same message, but in different generations
			int diff = writerIndex - receiver->readerIndex;

#ifndef abs
#define abs(x) (x>=0 ? x : -x)
#endif
			if(abs(diff) == msgCount) {
#undef abs
				return 0;
			}

			pos = writerIndex;

			return &pool[pos%msgCount];
			// if sender needs be granted access to the message, do it here
		}

	template <typename MSGTYPE, int msgCount>
		void SenderPort<MSGTYPE, msgCount>::sendMessage(MSGTYPE *message) {
		// in the simple case, does nothing
			writerIndex = (writerIndex + 1) % (2*msgCount);
		
		}

	template <typename MSGTYPE, int msgCount>
		MSGTYPE *ReceiverPort<MSGTYPE, msgCount>::receiveMessage() {
			// buffer empty?
			if(readerIndex == sender->writerIndex) {
				return 0;
			}

			return &sender->pool[readerIndex%msgCount];
			// if recipient needs write access to the message, do it here
		}

	template <typename MSGTYPE, int msgCount>
		void ReceiverPort<MSGTYPE, msgCount>::releaseMessage(MSGTYPE *message) {
			// if recipient was granted write access in receive, revoke it here

			// messages need be released in retrieved order
			readerIndex = (readerIndex+1) % (msgCount*2);
		}
}}; // namespace os::krn

#endif
