// $Id: TasklistTypes.h 995 2009-08-11 14:26:31Z mike $
// Created:   26.06.2006 lohmann

/** 
 * @file TasklistTypes.h
 * @brief Contains the class TasklistTypes which acts as a container for the Tasklist template.
 * 
 * A Tasklist is container of (Task, TaskPri) tuples, sorted by the integer type TaskPri. The Task
 * type (element type) is opaque for a Tasklist, it just has to be constructable and assignable.
 * Hence, a Tasklist just implements a sorting strategy. It is supposed to be used wherever
 * a priority queue for Task elements has to be maintained, such as the scheduler ready queue or the
 * waiting queue of a semaphore.
 * By providing additional information about the maximum number of elements as well as the maximum 
 * number of elements per priority, optimized implementations can be provided with partial 
 * specializations of the Tasklist template.
 * 
 */ 

#ifndef __TASKLISTTYPES__
#define __TASKLISTTYPES__

#include "stdinc.h"
 
namespace os { namespace krn {

struct TasklistTypesBase {
	
	// General Tasklist (NOT YET IMPLEMENTED)
	template< 
		class _Task, 			// ID type for a task. The Tasklist manages elements of this type.
		class _TaskPri, 		// Integer type for the priority of a task. 
		int _TASKS_MAX, 		// Maximum number of tasks this list has to manage
		int _TASKPRI_MAX,   	// Maximum number of priorities: 0 <= p < _TASKPRI_MAX [for each p instance of _TaskPri]
		int _PER_TASKPRI_MAX	// Maximum number of tasks this list has to manage for a single priority.
	> 
	struct Tasklist {

		// EMPTY, GENERAL CASE NOT YET IMPLEMENTED!	

	};

	// Specialization for max 1 task per priority class
	template< 	
		class _Task, 
		class _TaskPri, 
		int _TASKS_MAX, 
		int _TASKPRI_MAX 
		> 
	struct Tasklist< _Task, _TaskPri, _TASKS_MAX, _TASKPRI_MAX, /*>>>*/ 1 > {

		typedef _Task		Task;
		typedef _TaskPri	TaskPri;
		enum DOXY_SYMBOL(Consts) {
			TASKPRI_MAX		= _TASKPRI_MAX,
			TASKS_MAX		= _TASKS_MAX,
			PER_TASKPRI_MAX	= 1
		};
		typedef typename uint_t_minbits< TASKPRI_MAX >::T PriMap;

		volatile PriMap	enqueued_;
		Task	tasks_[ TASKPRI_MAX ];

	// Selectors
	
		//! Test for empty Tasklist
		Bool isEmpty() const;

		//! Returns the enqueued Task with the highest priority
		// isEmpty() == false: Result == Task with highest priority on list
		// isEmpty() == true : UNDEFINED
		Task head() const;
		
		//! Returns the highest priority of all enqueued tasks
		// (isEmpty() == false: Result == highest used priority on list 
		// (isEmpty() == true : UNDEFINED
		TaskPri headPri() const;

		//! Test element for on list membership
		// If Task task has been enqueued with TaskPri pri: Result = true
		Bool isEnqueued( Task task, TaskPri pri ) const;
		
	// Modifiers
	
		//! Initialitze Tasklist
		// after: isEmpty() == true
		void init();

		//! Insert element
		// isEnqueued(task, pri) == false: Result == task, isEnqueued(task, pri) == true
		// isEnqueued(task, pri) == true : UNDEFINED
		Task enqueue( Task task, TaskPri pri ); 
		
		//! Remove element
		// isEnqueued(task, pri) == true : Result == task, isEnqueued(task, pri) == false
		// isEnqueued(task, pri) == false: UNDEFINED
		Task remove( Task task, TaskPri currentPri );

		//! Remove element with highest priority
		// isEmpty() == false: [h=head(), p=headPri()]; Result == h, isEnqueued(h, p) == false
		// isEmpty() == true : UNDEFINED
		Task dequeue();
			
		//! Modifify priority of element on list
		// isEnqueued(task, currentPri) == true : Result = task, isEnqueued(task, newPri) == true, isEnqueued(task, currentPri) == false
		// isEnqueued(task, currentPri) == false: UNDEFINED	
		Task promote( Task task, TaskPri currentPri, TaskPri newPri );
		
	};
	
	// Specialization for N tasks with identical priority
	template< 	
		class _Task, 
		class _TaskPri, 
		int _TASKS_MAX
		> 
	struct Tasklist< _Task, _TaskPri, _TASKS_MAX, 0, _TASKS_MAX > {

		typedef _Task		Task;
		typedef _TaskPri	TaskPri;
		enum DOXY_SYMBOL(Consts) {
			TASKPRI_MAX		= 0,
			TASKS_MAX		= _TASKS_MAX,
			PER_TASKPRI_MAX	= _TASKS_MAX,
			INVALID_TASK    = _TASKS_MAX
		};
		typedef typename uint_t_minbits< TASKS_MAX >::T EnqueueMap;

		volatile EnqueueMap	enqueued_;
		Task	next_[ TASKS_MAX ];
		Task	head_;
		Task    tail_;

	// Selectors
	
		//! Test for empty Tasklist
		Bool isEmpty() const;

		//! Returns the enqueued Task with the highest priority
		// isEmpty() == false: Result == Task with highest priority on list
		// isEmpty() == true : UNDEFINED
		Task head() const;
		
		//! Returns the highest priority of all enqueued tasks
		// (isEmpty() == false: Result == highest used priority on list 
		// (isEmpty() == true : UNDEFINED
		TaskPri headPri() const;

		//! Test element for on list membership
		// If Task task has been enqueued with TaskPri pri: Result = true
		Bool isEnqueued( Task task, TaskPri pri ) const;
		
	// Modifiers
	
		//! Initialitze Tasklist
		// after: isEmpty() == true
		void init();

		//! Insert element
		// isEnqueued(task, pri) == false: Result == task, isEnqueued(task, pri) == true
		// isEnqueued(task, pri) == true : UNDEFINED
		Task enqueue( Task task, TaskPri pri ); 
		
		//! Remove element
		// isEnqueued(task, pri) == true : Result == task, isEnqueued(task, pri) == false
		// isEnqueued(task, pri) == false: UNDEFINED
		Task remove( Task task, TaskPri currentPri );

		//! Remove element with highest priority
		// isEmpty() == false: [h=head(), p=headPri()]; Result == h, isEnqueued(h, p) == false
		// isEmpty() == true : UNDEFINED
		Task dequeue();
			
		//! Modifify priority of element on list
		// isEnqueued(task, currentPri) == true : Result = task, isEnqueued(task, newPri) == true, isEnqueued(task, currentPri) == false
		// isEnqueued(task, currentPri) == false: UNDEFINED	
		Task promote( Task task, TaskPri currentPri, TaskPri newPri );
		
	};
	
};


// Aspects may override existing specialisations via introductions
struct TasklistTypes : public TasklistTypesBase {


};


//////////////////////////////////////////////////////////////////////////
// Implementation here, as PUMA doesn't like them inside the class :-(

template< class _Task, class _TaskPri, int _TASKS_MAX, int _TASKPRI_MAX > 
inline Bool TasklistTypesBase::Tasklist< _Task, _TaskPri, _TASKS_MAX, _TASKPRI_MAX, 1 >::isEmpty() const {
	return enqueued_ == 0;	 
}

template< class _Task, class _TaskPri, int _TASKS_MAX, int _TASKPRI_MAX > 
inline void TasklistTypesBase::Tasklist< _Task, _TaskPri, _TASKS_MAX, _TASKPRI_MAX, 1 >::init() {
	enqueued_ = 0;
	// Don't initialize array?
}

template< class _Task, class _TaskPri, int _TASKS_MAX, int _TASKPRI_MAX > 
inline Bool TasklistTypesBase::Tasklist< _Task, _TaskPri, _TASKS_MAX, _TASKPRI_MAX, 1 >::isEnqueued( Task task, TaskPri pri ) const {
	return testBit( enqueued_, pri ); // return ( enqueued_ & 1 << pri ) != 0; 
}

template< class _Task, class _TaskPri, int _TASKS_MAX, int _TASKPRI_MAX > 
inline _Task TasklistTypesBase::Tasklist< _Task, _TaskPri, _TASKS_MAX, _TASKPRI_MAX, 1 >::enqueue( Task task, TaskPri pri ) {
	tasks_[ pri ] = task;
	enqueued_ = setBit( enqueued_, pri ); // enqueued_ |= 1 << pri;
	return task;
}

template< class _Task, class _TaskPri, int _TASKS_MAX, int _TASKPRI_MAX > 
inline _Task TasklistTypesBase::Tasklist< _Task, _TaskPri, _TASKS_MAX, _TASKPRI_MAX, 1 >::head() const  {
	return tasks_[ highest1Bit( enqueued_ ) ];
}

template< class _Task, class _TaskPri, int _TASKS_MAX, int _TASKPRI_MAX > 
inline _TaskPri TasklistTypesBase::Tasklist< _Task, _TaskPri, _TASKS_MAX, _TASKPRI_MAX, 1 >::headPri() const  {
	return highest1Bit( enqueued_ );
}

template< class _Task, class _TaskPri, int _TASKS_MAX, int _TASKPRI_MAX > 
inline _Task TasklistTypesBase::Tasklist< _Task, _TaskPri, _TASKS_MAX, _TASKPRI_MAX, 1 >::remove( Task task, TaskPri currentPri ) {
	enqueued_ = clearBit( enqueued_, currentPri ); //enqueued_ &= ~(1 << currentPri);
	return task;	
}

template< class _Task, class _TaskPri, int _TASKS_MAX, int _TASKPRI_MAX > 
_Task TasklistTypesBase::Tasklist< _Task, _TaskPri, _TASKS_MAX, _TASKPRI_MAX, 1 >::dequeue() {
	return remove( head(), headPri() );
}
	
template< class _Task, class _TaskPri, int _TASKS_MAX, int _TASKPRI_MAX > 
_Task TasklistTypesBase::Tasklist< _Task, _TaskPri, _TASKS_MAX, _TASKPRI_MAX, 1 >::promote( Task task, TaskPri currentPri, TaskPri newPri ) {
	return enqueue( remove( task, currentPri ), newPri );
}

template< class _Task, class _TaskPri, int _TASKS_MAX> 
inline Bool TasklistTypesBase::Tasklist< _Task, _TaskPri, _TASKS_MAX, 0, _TASKS_MAX >::isEmpty() const {
	return enqueued_ == 0;	 
}

// effectively RR here ...

template< class _Task, class _TaskPri, int _TASKS_MAX > 
inline void TasklistTypesBase::Tasklist< _Task, _TaskPri, _TASKS_MAX, 0, _TASKS_MAX >::init() {
	enqueued_ = 0;
	head_ = (Task)INVALID_TASK;
	tail_ = (Task)INVALID_TASK;
}

template< class _Task, class _TaskPri, int _TASKS_MAX > 
inline Bool TasklistTypesBase::Tasklist< _Task, _TaskPri, _TASKS_MAX, 0, _TASKS_MAX >::isEnqueued( Task task, TaskPri pri ) const {
	return testBit( enqueued_, task ); // return ( enqueued_ & 1 << task ) != 0; 
}

template< class _Task, class _TaskPri, int _TASKS_MAX > 
inline _Task TasklistTypesBase::Tasklist< _Task, _TaskPri, _TASKS_MAX, 0, _TASKS_MAX >::enqueue( Task task, TaskPri pri ) {
	next_[ task ] = (Task)INVALID_TASK;
	if (tail_ == (Task)INVALID_TASK)
		head_ = task;
	else
		next_[tail_] = task;
	tail_ = task;
	enqueued_ = setBit( enqueued_, task ); // enqueued_ |= 1 << task;
	return task;
}

template< class _Task, class _TaskPri, int _TASKS_MAX > 
inline _Task TasklistTypesBase::Tasklist< _Task, _TaskPri, _TASKS_MAX, 0, _TASKS_MAX >::head() const  {
	return head_;
}

template< class _Task, class _TaskPri, int _TASKS_MAX > 
inline _TaskPri TasklistTypesBase::Tasklist< _Task, _TaskPri, _TASKS_MAX, 0, _TASKS_MAX >::headPri() const  {
	return 0;
}

template< class _Task, class _TaskPri, int _TASKS_MAX > 
inline _Task TasklistTypesBase::Tasklist< _Task, _TaskPri, _TASKS_MAX, 0, _TASKS_MAX >::remove( Task task, TaskPri currentPri ) {
	// find the previous task
	if (task == head_) {
		head_ = next_[task];
		if (head_ == (Task)INVALID_TASK)
			tail_ = (Task)INVALID_TASK;
	}
	else {
		Task prev = (Task)INVALID_TASK;
		for (Task t = head_; t != (Task)INVALID_TASK && next_[t] != task; t = next_[t])
			prev = t;
		if (prev != (Task)INVALID_TASK) {
			next_[prev] = next_[task];
			if (next_[prev] == (Task)INVALID_TASK)
				tail_ = prev;
		}
	}
	// clear the bit
	enqueued_ = clearBit( enqueued_, task ); //enqueued_ &= ~(1 << task);
	return task;	
}

template< class _Task, class _TaskPri, int _TASKS_MAX > 
_Task TasklistTypesBase::Tasklist< _Task, _TaskPri, _TASKS_MAX, 0, _TASKS_MAX >::dequeue() {
	Task result = head_;
	head_ = next_[head_];
	next_[result] = (Task)INVALID_TASK;
	if (head_ == (Task)INVALID_TASK)
		tail_ = (Task)INVALID_TASK;
   	// clear the bit
   	enqueued_ = clearBit( enqueued_, result ); //enqueued_ &= ~(1 << result);
	return result;
}
	
template< class _Task, class _TaskPri, int _TASKS_MAX > 
_Task TasklistTypesBase::Tasklist< _Task, _TaskPri, _TASKS_MAX, 0, _TASKS_MAX >::promote( Task task, TaskPri currentPri, TaskPri newPri ) {
	return task;
}


}} // namespace os::krn
#endif // __TASKLISTTYPES__


