// $Id: StackAllocator.h 995 2009-08-11 14:26:31Z mike $

#ifndef __StackAllocator_h__
#define __StackAllocator_h__

#include "util/types.h"
#include "util/defines.h"
#include "mem/BaseAllocator.h"
#include "util/container/Chain.h"

//! Allocates memory on the execution stack
template<class T, int S>
class StackAllocator : public BaseAllocator<T>{

// public typedefs and enums
public:
	//! Type definition for instatiated type
	typedef T value_t;
	//! Type definition for pointer to intantiated type
	typedef Ptr pointer_t;
	//! Type definition for this type of allocator
	typedef StackAllocator<T,S> alloc_t;
	
	enum{ SIZE=S };
	

// protected variables
protected:
	//! memory area for storing objects of \ref value_t
	UInt8 _pool[SIZE*sizeof(T)];
	
	//! array of pointers to objects of \ref value_t
	value_t* _ptr[SIZE];
	
	//! position in \ref _ptr pointing out the next free block of memory
	typename index_t_value<SIZE>::T _pos;
	
// public methods
public:
	//! constructor
	StackAllocator():_pos(0){
		for(int i=0;i<SIZE;++i){
			_ptr[i]=(value_t*) &(_pool[i*(sizeof(T))]);
		}
	}
	
	//! allocate memory area for one object
	/**
	 * \return adress to memory location
	 */
 	Ptr allocate();
 	
 	
 	//! allocate memory area for one object
	/**
	 * \return adress to memory location
	 */
 	Ptr allocate(size_t size);
 	
 	//! free memory of one object 
 	/**
 	 * \param obj is the adress of the object which should be freed
 	 */
	void free(Ptr obj);
};

template<class T, int SIZE>
Ptr StackAllocator<T,SIZE>::allocate(){
		if(_pos < SIZE){
		T* ret = _ptr[_pos];
		++_pos;
		return static_cast<Ptr>(ret);
		}
		return 0;
}

template<class T, int SIZE>
Ptr StackAllocator<T,SIZE>::allocate(size_t size){
		return allocate();
}
	
template<class T, int SIZE>
void StackAllocator<T,SIZE>::free(Ptr elem){
		--_pos;
		_ptr[_pos]=static_cast<value_t*>(elem);
}	



// fuer Chain<T> spezialisierte Variante, die die Struktur der zu verwaltenden 
// Objekte fuer sich selbst ausnutzt. (bringt nur Vorteile bei AVR und H8)

//! Specialized \ref StackAllocator for \ref Chain 
/**
 * This specialized implementation uses the Structure of the \ref Chain class 
 * in order to reduce the memory usage. 
 */
template<class T,int S>
class StackAllocator< Chain<T>,S> : public BaseAllocator<T> { 
  
// public typedefs and enums
public:
	
	//! Type definition for instatiated type
	typedef Chain<T> value_t;
	//! Type definition for pointer to intantiated type
	typedef Ptr pointer_t;
	//! Type definition for this type of allocator
	typedef StackAllocator< Chain< T > , S > alloc_t;
  	   
  	enum {SIZE=S};
  	   
  	/**
  	 * \brief Please document me
  	 */
  	template<class NT>
	struct rebind{
		typedef typename StackAllocator<NT,S>::alloc_t alloc_t;
	};   
  	
  	
// protected variables   
protected:
	//! pointer to the next free memory location
	value_t* _head;
	//! memory area for storing objects of \ref value_t
	UInt8 _pool[SIZE *sizeof(value_t)];
	
	// wrapper function which cast the uninitialized byte array to
	// the inititialized type. 
	// obsolete: increases code size
	/*
	inline Chain<T>* pool(int i){
		return (Chain<T>*) &(_pool[i*sizeof(value_t)]);
	}
	*/

public:
	
	//! constructor
	/**
	 * Initialize the structure of the memory pool by linking together the 
	 * \ref Chain objects.
	 */
	StackAllocator(){
		_head = (value_t*) _pool;
		for(int i=0;i<SIZE-1;++i){
			//pool(i)->setNext(pool(i+1));
			((value_t*) &_pool[i*sizeof(value_t)])->setNext((value_t*) &_pool[(i+1)*sizeof(value_t)]);
		}
	}
	
	//! TODO
	StackAllocator(const StackAllocator& alloc){
		STATIC_ERROR(copy_contructor_not_available_for_this_allocator);
	}
	
	//! TODO
	pointer_t allocate();
	
	//! TODO
	pointer_t allocate(int size);
	
	//! TODO
	void free(pointer_t elem);
};


template<class T, int S>
Ptr StackAllocator<Chain<T>,S>::allocate(){
		Ptr ret=static_cast<Ptr>(_head);
		if(_head != 0){
			_head=(Chain<T>*) _head->getNext();
		}
		return ret;
}

template<class T, int S>
Ptr StackAllocator<Chain<T>,S>::allocate(int size){
		return allocate();
}

template<class T, int S>
void StackAllocator<Chain<T>,S>::free(Ptr elem){
		(static_cast< Chain<T>* >(elem)).setNext(_head);
		_head=(static_cast< Chain<T>* >(elem));
}



#endif //__StackAllocator_h__
