#ifndef __TYPES_H__
#define __TYPES_H__

#define CHAR_BIT __CHAR_BIT__

// Int and UInt are the natural integer types.
// They are "natural" in the sense of the CPU, as they reflect the size of a CPU register.
// Note that sizeof(Int) == sizeof(int) does not hold on all architectures (especially 8-Bit),
// as gcc uses at least 16 bit for int types.

#ifdef __AVR__
// 8-Bit machines: Explicitly define as char

	typedef signed char		Int;
	typedef unsigned char	UInt;

#else
// all others: Default to built-in int

	typedef int				Int;
	typedef unsigned int	UInt;

#endif


namespace internal {
	class ERROR_inttype_not_available
	{
		ERROR_inttype_not_available();
	};

	template< bool FIRST, class A, class B > struct select
	{
		typedef A T;
	};
	template< class A, class B > struct select< false, A, B >
	{
		typedef B T;
	};

}	// namespace internal

/// Calculates the number of bytes (>=0) needed by an unsigned int
/// to hold 0 <= i < MAX_VAL. If MAX_VAL == 0, 0 is returned
template< unsigned long long MAX_VAL > struct bytes_needed_exact
{
	enum { RES = 1 + bytes_needed_exact< MAX_VAL / 256 >::RES };
};
template<> struct bytes_needed_exact< 0 >
{
	enum { RES = 0 };
};

/// Calculates the number of bytes (>=1) needed by an unsigned int
/// to hold 0 <= i < MAX_VAL. If MAX_VAL == 0, 1 is returned.
template< unsigned long long MAX_VAL > struct bytes_needed
{
	enum { RES = 1 + bytes_needed_exact< MAX_VAL / 256 >::RES };
};

/// Specialisation for 0
template<> struct bytes_needed< 0 >
{
	enum { RES = 1 };
};

template <typename T>
struct integer_is_signed {
  enum { RES = ( (T)~(T)0 < (T)0 ) };
};

// min< int A, int B >::RES
// Calculates the minimum of A and B
template< int A, int B > struct min {
	enum { RES = A < B ? A : B };
};

// max< int A, int B >::RES
// Calculates the maximum of A and B
template< int A, int B > struct max {
	enum { RES = A > B ? A : B };
};

// Retrieves a signed/unsigned integer type with sizeof( T ) == BYTES
template< unsigned BYTES, bool SIGNED >
struct int_t
{
	typedef typename internal::select< sizeof( signed char ) * CHAR_BIT == 8 * BYTES,  signed char,
		    typename internal::select< sizeof( signed short ) * CHAR_BIT == 8 * BYTES, signed short,
			typename internal::select< sizeof( signed int ) * CHAR_BIT == 8 * BYTES, signed int,
			typename internal::select< sizeof( signed long ) * CHAR_BIT == 8 * BYTES, signed long,
			typename internal::select< sizeof( signed long long ) * CHAR_BIT == 8 * BYTES, signed long long,
			internal::ERROR_inttype_not_available >::T >::T >::T >::T >::T T;
};
template< unsigned BYTES >
struct int_t< BYTES, false>
{
	typedef typename internal::select< sizeof( unsigned char ) * CHAR_BIT == 8 * BYTES, unsigned char,
		    typename internal::select< sizeof( unsigned short ) * CHAR_BIT == 8 * BYTES, unsigned short,
			typename internal::select< sizeof( unsigned int ) * CHAR_BIT == 8 * BYTES, unsigned int,
			typename internal::select< sizeof( unsigned long ) * CHAR_BIT == 8 * BYTES, unsigned long,
			typename internal::select< sizeof( unsigned long long ) * CHAR_BIT == 8 * BYTES, unsigned long long,
			internal::ERROR_inttype_not_available >::T >::T >::T >::T >::T T;
};

// Retrieves the smallest unsigned integer type with sizeof( T ) >= BYTES
template< unsigned BYTES >
struct uint_t_min
{
	typedef typename internal::select< sizeof( unsigned char ) * CHAR_BIT >= 8 * BYTES, unsigned char,
		    typename internal::select< sizeof( unsigned short ) * CHAR_BIT >= 8 * BYTES, unsigned short,
			typename internal::select< sizeof( unsigned int ) * CHAR_BIT >= 8 * BYTES, unsigned int,
			typename internal::select< sizeof( unsigned long ) * CHAR_BIT >= 8 * BYTES, unsigned long,
			typename internal::select< sizeof( unsigned long long ) * CHAR_BIT >= 8 * BYTES, unsigned long long,
			internal::ERROR_inttype_not_available >::T >::T >::T >::T >::T T;
};


// Retrieves the smallest unsigned integer type with sizeof( T ) >= BYTES
template< unsigned BYTES >
struct int_t_min
{
	typedef typename internal::select< sizeof( char ) * CHAR_BIT >= 8 * BYTES, signed char,
		    typename internal::select< sizeof( short ) * CHAR_BIT >= 8 * BYTES, signed short,
			typename internal::select< sizeof( int ) * CHAR_BIT >= 8 * BYTES, signed int,
			typename internal::select< sizeof( long ) * CHAR_BIT >= 8 * BYTES, signed long,
			typename internal::select< sizeof( long long ) * CHAR_BIT >= 8 * BYTES, signed long long,
			internal::ERROR_inttype_not_available >::T >::T >::T >::T >::T T;
};

// Retrieves the smallest unsigned integer type that is able to
// hold 0 <= i < MAXVAL
template< unsigned long long MAXVAL >
struct uint_t_value
{
	typedef typename uint_t_min< bytes_needed< MAXVAL >::RES >::T T;
};

// Retrieves the smallest signed integer type that is able to
// hold -MAXVAL < i < MAXVAL.
// If the archtecture has two's complement arithmetics, then the
// constraint -MAXVAL <= i < MAXVAL is also valid.
template< signed long long MAXVAL >
struct int_t_value
{
	typedef typename internal::select<MAXVAL >= 0,
					typename int_t_min< bytes_needed< ((unsigned long long)MAXVAL) * 2 - 1>::RES >::T,
					internal::ERROR_inttype_not_available >::T T;
};

// Retrieves the *optimal* unsigned type that is able to
// hold: 0 <= i < MAXVAL. The returned type T is optimal with respect to code size/speed
// by extending it to at least the size of a register: sizeof(T) >= sizeof(UInt)
template< unsigned long long MAXVAL >
struct uint_t_value_opt {
	typedef typename uint_t_min< max< bytes_needed< MAXVAL >::RES, sizeof( UInt ) >::RES >::T T;
};

// Retrieves the smallest unsigned integer type that is able to hold BITS bits
template< char BITS >
struct uint_t_minbits
{
//	typedef typename uint_t_min< bytes_needed< (BITS / CHAR_BIT) + ( (BITS % CHAR_BIT) != 0 ) >::RES >::T T;
	typedef typename uint_t_min< (BITS / CHAR_BIT) + ( (BITS % CHAR_BIT) != 0 ) >::T T;
};

// Retrieves the *optimal* unsigned integer type that is able to hold BITS bits
// The returned type T is optimal with respect to code size/speed
// by extending it to at least the size of a register: sizeof(T) >= sizeof(UInt)
template< char BITS >
struct uint_t_minbits_opt
{
	typedef typename uint_t_min< max< bytes_needed<
			(BITS / CHAR_BIT) + ( (BITS % CHAR_BIT) != 0 ) >::RES,
			sizeof( UInt ) >::RES
	>::T T;
};

template<class INT_TYPE>
struct int_t_to_unsigned {
	typedef typename int_t<sizeof(INT_TYPE), false>::T T;
};

template<class INT_TYPE>
struct int_t_to_signed {
	typedef typename int_t<sizeof(INT_TYPE), true>::T T;
};

// Machine independent definition of sized integer types
typedef int_t< 1, true >::T		Int8;
typedef int_t< 2, true >::T		Int16;
typedef int_t< 4, true >::T		Int32;
typedef int_t< 8, true >::T		Int64;
typedef int_t< 1, false >::T	UInt8;
typedef int_t< 2, false >::T	UInt16;
typedef int_t< 4, false >::T	UInt32;
typedef int_t< 8, false >::T	UInt64;

typedef uint_t_minbits_opt<8>::T UInt8Opt;
typedef uint_t_minbits_opt<16>::T UInt16Opt;
typedef uint_t_minbits_opt<32>::T UInt32Opt;

// Other standard types
typedef char				Char;
typedef void*				Ptr;
// two following merged with stddef.h
//typedef __SIZE_TYPE__ 		size_t;
//typedef __PTRDIFF_TYPE__ 	ptrdiff_t;
typedef Int					Bool;

// Integer types that have exactly the size of a pointer
typedef int_t< sizeof ( Ptr ), true >::T	IntPtr;
typedef int_t< sizeof ( Ptr ), false >::T	UIntPtr;

template<class INT_TYPE>
struct int_t_is_signed {
  static const INT_TYPE zeros = (INT_TYPE)0;
  static const INT_TYPE one_maxshifted = ((INT_TYPE)1) << (8 * sizeof(INT_TYPE) - 1);
  static const bool RES = one_maxshifted < zeros;
};

namespace internal {
	struct std_not_zero {
		template< class T > static Bool not_zero( T val ) {
			return val != 0;
		}
	};
	struct opt_not_zero {
		template< class T > static Bool not_zero( T val ) {
			return (Bool) val;
		}
	};
}

// Efficient !=0 test for any type, result returned as Bool.
// Optimized by simply casting to Bool if type is loss-less convertible,
// otherwise a real test against 0 is performed.
template< class T > inline Bool not_zero( T val ) {
	typedef typename internal::select< sizeof( T ) <= sizeof( Bool ),
		internal::std_not_zero, internal::opt_not_zero >::T Impl;
	return Impl::not_zero(val);
}


// Now follows the definition of types and helpers
// that might be defined machine-specific with respect to
// optimal results. Therefore include the architecture-specific
// header first.

#include "types_arch.h"
#include "assert.h"


// intex_t_value< MAXVAL >::T
//
// Retrieves the *optimal* unsinged type
// to hold an index ranging 0 <= i < MAXVAL
//
// The resulting type is optimal in terms of code size
#ifndef __TYPES_intex_t_value
	template< unsigned long long MAXVAL=16 >
	struct index_t_value : public uint_t_value< MAXVAL > {};
#endif

// Conversion functions and helper
// might be considered plattform-specific one day, but not yet.
inline UInt32 upperHalf( UInt64 n ) {
	return (UInt32) (n >> 32);
}
inline UInt32 lowerHalf( UInt64 n ) {
	return (UInt32) (n & 0xffffffff);
}
inline UInt16 upperHalf( UInt32 n ) {
	return (UInt16) (n >> 16);
}
inline UInt16 lowerHalf( UInt32 n ) {
	return (UInt16) (n & 0xffff);
}
inline UInt8 upperHalf( UInt16 n ) {
	return (UInt8) (n >> 8);
}
inline UInt8 lowerHalf( UInt16 n ) {
	return (UInt8) (n & 0xff);
}

// returns the index of the most significant bit that is 1
#ifndef __TYPES_highest1Bit
	template< class INT_TYPE > inline UInt highest1Bit( INT_TYPE n ) {
		STATIC_ASSERT(sizeof(INT_TYPE) > 0, higest1Bit_not_implemented);
		return 0;
	}
#endif

#ifndef __TYPES_clearBit
template< class INT_TYPE > inline INT_TYPE clearBit(INT_TYPE val, UInt32 bit_index ) {
	return val & ~(1 << bit_index);
}
#endif


#ifndef __TYPES_setBit
template< class INT_TYPE > inline INT_TYPE setBit(INT_TYPE val, UInt32 bit_index ) {
	return val | (1 << bit_index);
}
#endif

#ifndef __TYPES_testBit
template< class INT_TYPE > inline Bool testBit(INT_TYPE val, UInt32 bit_index ) {
	return not_zero( val & (1 << bit_index) );
}
#endif



#endif // __TYPES_H__

