/*****************************************************************************
 * CiAO Proof-of-Concept: instantiation
 * 
 * This example demonstrates, I) how singelton instances can be maintained in a 
 * layered system that follows the principle of minimal extension. It 
 * demonstrates furthermore, II) how the CiAO layers interact with each other
 * and how the actual ciao::XXX classes are constructed.  
 * 
 * I)
 * Given a system that consists of multiple layers. The bottom layer 0
 * defines driver-classes for various devices (e.g. AVRUART0), each is maintained
 * as a singleton (instance returned by a static method "Inst()", e.g.  
 * AVRUART0& AVURART0::Inst()). Optional additional layers define extensions
 * to the driver-class (e.g. layer 2 with Serial0) by inheritance.
 * 
 * Problem:  The one and only singleton instance should now be of the refined
 *           type (here os::dev::std::Serial0). This means that the original
 *           instance must be "overridden". (hw::dev::AVRUART0::Inst() must return
 *           the same instance as os::dev::std::Serial0, which is of type 
 *           ...::Serial0)
 * 
 * Solution: Each refinement defines a corresponding "XXX_Inst" aspect.
 *           The aspect gives around advice to the base class' Inst() method
 *           to return the refined instance instead of the original one. As the
 *           old Inst() method was the only place where the 
 *           original instance was actually referenced, it completely disappears 
 *           due to symbol-level linking.
 *              
 *   -------------------------------------------------------------------
 *   3: ciao::dev::std                            
 *                      Serial0 : public hw::hal::Serial0       config 1
 *                              | public os::dev:std::Serial0   config 2 
 * 
 *   -------------------------------------------------------------------
 *   2: os::dev::std                            
 *                      Serial0 : public hw::hal::Serial0  
 * 
 *   -------------------------------------------------------------------
 *   1: hw::hal                            
 *                      Serial0 : alias hw::dev::AVRUART0  
 * 
 *   -------------------------------------------------------------------
 *   0: hw::dev                            
 *                      AVRUART0  
 * 
 *   -------------------------------------------------------------------
 *
 * II) 
 * Depending on the configuration, an actual "Serial0" class may
 * be either refined in layer 2 or not. The ciao layer (layer 3), however,
 * always reflects the highest configured variant, that is os::dev::std::Serial0
 * if configured, otherwise hw::hal::Serial0 or now Serial0 class at all.
 *   
 * Classes on the ciao-layer are just thin wrappers around their actual
 * implementation classes. Their aim is to be able to give advice for transitions
 * from user code to OS code, therefore every (public) method of the implementation
 * class has to be replicated into the corresponding ciao class. On the long
 * term, the ciao-classes should be generated by a tool. At time of this writing
 * they have, however, to be written by hand. This also means, that the 
 * actual implementation class has to be configured explicitly. The concept study 
 * demonstrates, how a hand-written ciao class does look like.
 * 
 * The following flags are relevant for configuring the ciao- (and partly also 
 * the os-) layer. They are generated by pure::variance and applied by a forced
 * include:
 * 
 *   // general flags
 * 
 *   cfCIAO_DECOUPLE:  decouple the address area of the ciao instances from the
 *                     the address area os and/or hw instances are located.
 *                     The first should be placed in the "application address
 *                     space", the latter in the "kernel address space".
 *                     This basically means that ciao-instances must *not* 
 *                     replace the singletons of lower layers, but delegate
 *                     to them.
 * 
 *   cfCIAO_INLINE:    inline the ciao-code into the application. Only
 *                     suitable if cfCIAO_DECOUPLED is not defined.
 * 
 *   cfOS_INLINE:      inline the os-code into the ciao-layer
 * 
 * 
 *   // ciao-layer component implementation flags
 * 
 *   cfCIAO_SERIAL0_IMPL_HW: if defined, ciao::dev::std::Serial0 should be 
 *                           implemented by hw::hal::Serial0, otherwise
 *                           it is implemented by os::dev::std::Serial0
 *    
 *****************************************************************************/
 
 
#include <stdio.h>
#include "ciao/dev/std/Serial0.h"
#include "hw/dev/AVRUART0.h"

__attribute__ ((noinline))
hw::dev::AVRUART0& getInst() {
	asm("nop");
	return hw::dev::AVRUART0::Inst();
	asm("nop");
}

int main(){
	ciao::dev::std::Serial0& ser = ciao::dev::std::Serial0::Inst();
	
// send(char*) is only defined by os::dev::std::Serial0, but
// not by hw::hal::Serial0
#ifndef cfCIAO_SERIAL0_IMPL_HW
	ser.send( "Hello, World" );
#endif
	ser.send( '!' );
	printf("Instance check:\n");
	printf("&ciao::dev::std::Serial0::Inst() == &hw::dev::AVRUART0::Inst(): %d\n", (void*)&ciao::dev::std::Serial0::Inst() == (void*)&getInst() );
	
}
